@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2018 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2019 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -49,9 +49,10 @@
 #define IF_ACTIVE	1
 #define IF_ACTIVE_USER	2
 
-#define LINK_UP		1
-#define LINK_UNKNOWN	0
-#define LINK_DOWN	-1
+#define	LINK_UP		1
+#define	LINK_UNKNOWN	0
+#define	LINK_DOWN	-1
+#define	LINK_DOWN_IFFUP	-2
 
 #define IF_DATA_IPV4	0
 #define IF_DATA_ARP	1
@@ -84,8 +85,9 @@ struct interface {
 	unsigned short vlanid;
 	unsigned int metric;
 	int carrier;
-	int wireless;
-	uint8_t ssid[IF_SSIDLEN + 1]; /* NULL terminated */
+	bool media_valid;
+	bool wireless;
+	uint8_t ssid[IF_SSIDLEN];
 	unsigned int ssid_len;
 
 	char profile[PROFILE_LEN];
@@ -140,14 +142,10 @@ struct dhcpcd_ctx {
 	struct rt_head froutes;	/* free routes for re-use */
 
 	int pf_inet_fd;
-#ifdef IFLR_ACTIVE
-	int pf_link_fd;
-#endif
 	void *priv;
 	int link_fd;
 	int seq;	/* route message sequence no */
 	int sseq;	/* successful seq no sent */
-	struct iovec iov[1];	/* generic iovec buffer */
 
 #ifdef USE_SIGNALS
 	sigset_t sigset;
@@ -166,13 +164,6 @@ struct dhcpcd_ctx {
 
 	char *randomstate; /* original state */
 
-	/* Used to track the last routing message,
-	 * so we can ignore messages the parent process sent
-	 * but the child receives when forking.
-	 * getppid(2) is unreliable because we detach. */
-	pid_t ppid;	/* parent pid */
-	int pseq;	/* last seq in parent */
-
 #ifdef INET
 	struct dhcp_opt *dhcp_opts;
 	size_t dhcp_opts_len;
@@ -189,15 +180,6 @@ struct dhcpcd_ctx {
 	uint8_t *secret;
 	size_t secret_len;
 
-	unsigned char ctlbuf[IP6BUFLEN];
-	struct sockaddr_in6 from;
-	struct msghdr sndhdr;
-	struct iovec sndiov[1];
-	unsigned char sndbuf[CMSG_SPACE(sizeof(struct in6_pktinfo))];
-	struct msghdr rcvhdr;
-	char ntopbuf[INET6_ADDRSTRLEN];
-	const char *sfrom;
-
 	int nd_fd;
 	struct ra_head *ra_routers;
 
@@ -205,8 +187,10 @@ struct dhcpcd_ctx {
 
 	struct dhcp_opt *nd_opts;
 	size_t nd_opts_len;
+#ifdef DHCP6
 	struct dhcp_opt *dhcp6_opts;
 	size_t dhcp6_opts_len;
+#endif
 
 #ifndef __linux__
 	int ra_global;
@@ -230,6 +214,7 @@ int dhcpcd_ifafwaiting(const struct interface *);
 int dhcpcd_afwaiting(const struct dhcpcd_ctx *);
 pid_t dhcpcd_daemonise(struct dhcpcd_ctx *);
 
+void dhcpcd_linkoverflow(struct dhcpcd_ctx *);
 int dhcpcd_handleargs(struct dhcpcd_ctx *, struct fd_list *, int, char **);
 void dhcpcd_handlecarrier(struct dhcpcd_ctx *, int, unsigned int, const char *);
 int dhcpcd_handleinterface(void *, int, const char *);
