@@ -1,6 +1,6 @@
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2018 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2019 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -78,12 +78,20 @@ if_free(struct interface *ifp)
 
 	if (ifp == NULL)
 		return;
+#ifdef IPV4LL
 	ipv4ll_free(ifp);
+#endif
+#ifdef INET
 	dhcp_free(ifp);
 	ipv4_free(ifp);
+#endif
+#ifdef DHCP6
 	dhcp6_free(ifp);
+#endif
+#ifdef INET6
 	ipv6nd_free(ifp);
 	ipv6_free(ifp);
+#endif
 	rt_freeif(ifp);
 	free_options(ifp->ctx, ifp->options);
 	free(ifp);
@@ -101,12 +109,6 @@ if_opensockets(struct dhcpcd_ctx *ctx)
 	if (ctx->pf_inet_fd == -1)
 		return -1;
 
-#ifdef IFLR_ACTIVE
-	ctx->pf_link_fd = xsocket(PF_LINK, SOCK_DGRAM | SOCK_CLOEXEC, 0);
-	if (ctx->pf_link_fd == -1)
-		return -1;
-#endif
-
 	return 0;
 }
 
@@ -116,10 +118,6 @@ if_closesockets(struct dhcpcd_ctx *ctx)
 
 	if (ctx->pf_inet_fd != -1)
 		close(ctx->pf_inet_fd);
-#ifdef IFLR_ACTIVE
-	if (ctx->pf_link_fd != -1)
-		close(ctx->pf_link_fd);
-#endif
 
 	if (ctx->priv) {
 		if_closesockets_os(ctx);
@@ -138,21 +136,32 @@ if_carrier(struct interface *ifp)
 
 	memset(&ifr, 0, sizeof(ifr));
 	strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
-	if (ioctl(ifp->ctx->pf_inet_fd, SIOCGIFFLAGS, &ifr) == -1)
+	r = ioctl(ifp->ctx->pf_inet_fd, SIOCGIFFLAGS, &ifr);
+	if (r != -1)
+		ifp->flags = (unsigned int)ifr.ifr_flags;
+
+#ifdef __sun
+	return if_carrier_os(ifp);
+#else
+	if (r == -1)
 		return LINK_UNKNOWN;
-	ifp->flags = (unsigned int)ifr.ifr_flags;
 
 #ifdef SIOCGIFMEDIA
 	memset(&ifmr, 0, sizeof(ifmr));
 	strlcpy(ifmr.ifm_name, ifp->name, sizeof(ifmr.ifm_name));
 	if (ioctl(ifp->ctx->pf_inet_fd, SIOCGIFMEDIA, &ifmr) != -1 &&
 	    ifmr.ifm_status & IFM_AVALID)
+	{
+		ifp->media_valid = true;
 		r = (ifmr.ifm_status & IFM_ACTIVE) ? LINK_UP : LINK_DOWN;
-	else
+	} else {
+		ifp->media_valid = false;
 		r = ifr.ifr_flags & IFF_RUNNING ? LINK_UP : LINK_UNKNOWN;
+	}
 #else
 	r = ifr.ifr_flags & IFF_RUNNING ? LINK_UP : LINK_DOWN;
 #endif
+#endif /* __sun */
 	return r;
 }
 
@@ -240,10 +249,8 @@ if_learnaddrs(struct dhcpcd_ctx *ctx, struct if_head *ifs,
 			addrflags = if_addrflags(ifp, &addr->sin_addr,
 			    ifa->ifa_name);
 			if (addrflags == -1) {
-				if (errno != EEXIST)
-					logerr("%s: if_addrflags: %s",
-					    __func__,
-					    inet_ntoa(addr->sin_addr));
+				if (errno != EEXIST && errno != EADDRNOTAVAIL)
+					logerr("%s: if_addrflags", __func__);
 				continue;
 			}
 #endif
@@ -266,7 +273,7 @@ if_learnaddrs(struct dhcpcd_ctx *ctx, struct if_head *ifs,
 			addrflags = if_addrflags6(ifp, &sin6->sin6_addr,
 			    ifa->ifa_name);
 			if (addrflags == -1) {
-				if (errno != EEXIST)
+				if (errno != EEXIST && errno != EADDRNOTAVAIL)
 					logerr("%s: if_addrflags6", __func__);
 				continue;
 			}
@@ -332,12 +339,10 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 	struct ifreq ifr;
 #endif
 #ifdef IFLR_ACTIVE
-	struct if_laddrreq iflr;
+	struct if_laddrreq iflr = { .flags = IFLR_PREFIX };
+	int link_fd;
 #endif
 
-#ifdef IFLR_ACTIVE
-	memset(&iflr, 0, sizeof(iflr));
-#endif
 #elif AF_PACKET
 	const struct sockaddr_ll *sll;
 #endif
@@ -346,11 +351,21 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 		logerr(__func__);
 		return NULL;
 	}
-	TAILQ_INIT(ifs);
 	if (getifaddrs(ifaddrs) == -1) {
 		logerr(__func__);
-		goto out;
+		free(ifs);
+		return NULL;
 	}
+	TAILQ_INIT(ifs);
+
+#ifdef IFLR_ACTIVE
+	link_fd = xsocket(PF_LINK, SOCK_DGRAM | SOCK_CLOEXEC, 0);
+	if (link_fd == -1) {
+		logerr(__func__);
+		free(ifs);
+		return NULL;
+	}
+#endif
 
 	for (ifa = *ifaddrs; ifa; ifa = ifa->ifa_next) {
 		if (ifa->ifa_addr != NULL) {
@@ -447,7 +462,7 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 			    MIN(ifa->ifa_addr->sa_len, sizeof(iflr.addr)));
 			iflr.flags = IFLR_PREFIX;
 			iflr.prefixlen = (unsigned int)sdl->sdl_alen * NBBY;
-			if (ioctl(ctx->pf_link_fd, SIOCGLIFADDR, &iflr) == -1 ||
+			if (ioctl(link_fd, SIOCGLIFADDR, &iflr) == -1 ||
 			    !(iflr.flags & IFLR_ACTIVE))
 			{
 				if_free(ifp);
@@ -464,7 +479,7 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 			case IFT_PPP: /* FALLTHROUGH */
 #endif
 #ifdef IFT_PROPVIRTUAL
-			case IFT_PROPVIRTUAL: /* FALLTHROUGH */
+			case IFT_PROPVIRTUAL:
 #endif
 #if defined(IFT_BRIDGE) || defined(IFT_PPP) || defined(IFT_PROPVIRTUAL)
 				/* Don't allow unless explicit */
@@ -478,6 +493,7 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 					    ifp->name);
 					active = IF_INACTIVE;
 				}
+				__fallthrough; /* Appease gcc-7 */
 				/* FALLTHROUGH */
 #endif
 #ifdef IFT_L2VLAN
@@ -586,7 +602,7 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 		 * we can work them out. */
 		ifp->metric = 200 + ifp->index;
 		if (if_getssid(ifp) != -1) {
-			ifp->wireless = 1;
+			ifp->wireless = true;
 			ifp->metric += 100;
 		}
 #endif
@@ -599,7 +615,9 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 		TAILQ_INSERT_TAIL(ifs, ifp, next);
 	}
 
-out:
+#ifdef IFLR_ACTIVE
+	close(link_fd);
+#endif
 	return ifs;
 }
 
@@ -700,6 +718,11 @@ if_domtu(const struct interface *ifp, short int mtu)
 	int r;
 	struct ifreq ifr;
 
+#ifdef __sun
+	if (mtu == 0)
+		return if_mtu_os(ifp);
+#endif
+
 	memset(&ifr, 0, sizeof(ifr));
 	strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
 	ifr.ifr_mtu = mtu;
@@ -728,19 +751,24 @@ if_cmp(const struct interface *si, const struct interface *ti)
 		return -1;
 	if (si->carrier < ti->carrier)
 		return 1;
-
+#ifdef INET
 	if (D_STATE_RUNNING(si) && !D_STATE_RUNNING(ti))
 		return -1;
 	if (!D_STATE_RUNNING(si) && D_STATE_RUNNING(ti))
 		return 1;
+#endif
+#ifdef INET6
 	if (RS_STATE_RUNNING(si) && !RS_STATE_RUNNING(ti))
 		return -1;
 	if (!RS_STATE_RUNNING(si) && RS_STATE_RUNNING(ti))
 		return 1;
+#endif
+#ifdef DHCP6
 	if (D6_STATE_RUNNING(si) && !D6_STATE_RUNNING(ti))
 		return -1;
 	if (!D6_STATE_RUNNING(si) && D6_STATE_RUNNING(ti))
 		return 1;
+#endif
 
 #ifdef INET
 	/* Special attention needed here due to states and IPv4LL. */
@@ -756,6 +784,17 @@ if_cmp(const struct interface *si, const struct interface *ti)
 	return 0;
 }
 
+#ifdef ALIAS_ADDR
+int
+if_makealias(char *alias, size_t alias_len, const char *ifname, int lun)
+{
+
+	if (lun == 0)
+		return strlcpy(alias, ifname, alias_len);
+	return snprintf(alias, alias_len, "%s:%u", ifname, lun);
+}
+#endif
+
 /* Sort the interfaces into a preferred order - best first, worst last. */
 void
 if_sortinterfaces(struct dhcpcd_ctx *ctx)
@@ -785,6 +824,68 @@ if_sortinterfaces(struct dhcpcd_ctx *ctx)
 	TAILQ_CONCAT(ctx->ifaces, &sorted, next);
 }
 
+struct interface *
+if_findifpfromcmsg(struct dhcpcd_ctx *ctx, struct msghdr *msg, int *hoplimit)
+{
+	struct cmsghdr *cm;
+	unsigned int ifindex = 0;
+	struct interface *ifp;
+#if defined(INET) && defined(IP_PKTINFO)
+	struct in_pktinfo ipi;
+#endif
+#ifdef INET6
+	struct in6_pktinfo ipi6;
+#else
+	UNUSED(hoplimit);
+#endif
+
+	for (cm = (struct cmsghdr *)CMSG_FIRSTHDR(msg);
+	     cm;
+	     cm = (struct cmsghdr *)CMSG_NXTHDR(msg, cm))
+	{
+#if defined(INET) && defined(IP_PKTINFO)
+		if (cm->cmsg_level == IPPROTO_IP) {
+			switch(cm->cmsg_type) {
+			case IP_PKTINFO:
+				if (cm->cmsg_len != CMSG_LEN(sizeof(ipi)))
+					continue;
+				memcpy(&ipi, CMSG_DATA(cm), sizeof(ipi));
+				ifindex = (unsigned int)ipi.ipi_ifindex;
+				break;
+			}
+		}
+#endif
+#ifdef INET6
+		if (cm->cmsg_level == IPPROTO_IPV6) {
+			switch(cm->cmsg_type) {
+			case IPV6_PKTINFO:
+				if (cm->cmsg_len != CMSG_LEN(sizeof(ipi6)))
+					continue;
+				memcpy(&ipi6, CMSG_DATA(cm), sizeof(ipi6));
+				ifindex = (unsigned int)ipi6.ipi6_ifindex;
+				break;
+			case IPV6_HOPLIMIT:
+				if (cm->cmsg_len != CMSG_LEN(sizeof(int)))
+					continue;
+				if (hoplimit == NULL)
+					break;
+				memcpy(hoplimit, CMSG_DATA(cm), sizeof(int));
+				break;
+			}
+		}
+#endif
+	}
+
+	/* Find the receiving interface */
+	TAILQ_FOREACH(ifp, ctx->ifaces, next) {
+		if (ifp->index == ifindex)
+			break;
+	}
+	if (ifp == NULL)
+		errno = ESRCH;
+	return ifp;
+}
+
 int
 xsocket(int domain, int type, int protocol)
 {
@@ -792,6 +893,9 @@ xsocket(int domain, int type, int protocol)
 #if !defined(HAVE_SOCK_CLOEXEC) || !defined(HAVE_SOCK_NONBLOCK)
 	int xflags, xtype = type;
 #endif
+#ifdef SO_RERROR
+	int on;
+#endif
 
 #ifndef HAVE_SOCK_CLOEXEC
 	if (xtype & SOCK_CLOEXEC)
@@ -816,6 +920,13 @@ xsocket(int domain, int type, int protocol)
 		goto out;
 #endif
 
+#ifdef SO_RERROR
+	/* Tell recvmsg(2) to return ENOBUFS if the receiving socket overflows. */
+	on = 1;
+	if (setsockopt(s, SOL_SOCKET, SO_RERROR, &on, sizeof(on)) == -1)
+		logerr("%s: SO_RERROR", __func__);
+#endif
+
 	return s;
 
 #if !defined(HAVE_SOCK_CLOEXEC) || !defined(HAVE_SOCK_NONBLOCK)
